/*
    libporty - IPv4 UDP
    Copyright (C) 2005 gpmi-extension project
    Copyright (C) 2010..2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Contact (email and IRC): http://igor2.repo.hu/contact.html
*/

#include <string.h>
#include "udp4.h"
#include "dns4.h"

/**
 * Opens a listen-socket on UDP for IPv4.
 *
 * @param socket The socket that will get the listener
 * @param hostname The hostname (may be an IP) to bind to
 * @param port The bind-port
 * @param non_blocking Bitfield P_nonblock_nonblock_operation
 * @param broadcast
 * @return see enum P_net_err
 */
int P_udp4_listen(P_net_socket *sock, const char *loc_hostname, int loc_port, P_net_nonblock_t non_blocking, P_net_broadcast_t broadcast)
{
	P_ipv4_addr_t ip;
	struct sockaddr_in sin;

	if (P_dns4_name_to_IP(&ip, loc_hostname) != 0)
		return P_net_err_dns;

	*sock = P_socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (*sock == P_NET_INVALID_SOCKET)
		return P_net_err_socket;

	/* Switch to non-blocking, if needed */
	if (non_blocking & P_net_nonblock_socket) {
		if (P_net_set_nonblocking(*sock) != 0 && (non_blocking & P_net_nonblock_quit_on_error)) {
			P_net_close(*sock);
			return P_net_err_nonblock;
		}
	}

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = ip;
	sin.sin_port = htons(loc_port);

	/* Bind */
	if (bind(*sock, (struct sockaddr *)&sin, sizeof(sin)) != 0) {
		P_net_close(*sock);
		return P_net_err_bind;
	}

	/* Check for broadcast */
	if (broadcast & P_net_broadcast_socket) {
		if (P_net_enable_broadcast(*sock) != 0 && (broadcast & P_net_broadcast_quit_on_error)) {
			P_net_close(*sock);
			return P_net_err_broadcast;
		}
	}

	return 0;
}

/**
 * Connect to a hostname / port.
 *
 * @param sock The socket which will connect
 * @param hostname The host to connect to
 * @param port The port to connect to
 * @param non_blocking Bitfield, enum P_nonblock

  Bit 1: quit on error, bit 2: connect non-blocking (watch out with this!),
 *           bit 4: set non-blocking after connect
 * @return see enum P_net_err
 */
int P_udp4_connect(P_net_socket *sock, const char *rem_hostname, int rem_port, const char *loc_hostname, int loc_port, P_net_nonblock_t non_blocking)
{
	P_ipv4_addr_t ip, loc_ip;
	struct sockaddr_in sin;
	int r, x = 1;
	unsigned char *ip_ = (unsigned char *)&ip;

	if (P_dns4_name_to_IP(&ip, rem_hostname) != 0)
		return P_net_err_dns;

	*sock = P_socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (*sock == P_NET_INVALID_SOCKET)
		return P_net_err_socket;

#warning this is not good enough - byte order!
	if (*ip_ == 255) {
		r = setsockopt (*sock, SOL_SOCKET, SO_BROADCAST, &x, sizeof (x));
		if ((r != 0) && (non_blocking & P_net_nonblock_quit_on_error))
			return P_net_err_broadcast;
	}
	/* Switch to non-blocking, if needed */
	if (non_blocking & P_net_nonblock_connect) {
		if (P_net_set_nonblocking(*sock) != 0 && (non_blocking & P_net_nonblock_quit_on_error)) {
			P_net_close(*sock);
			return P_net_err_nonblock;
		}
	}

	/* set up sin for local and remote host/port */
	memset((char *) &sin, 0, sizeof(struct sockaddr_in));
	sin.sin_family = AF_INET;

#warning TODO: local binding is needed even if loc_hostname is NULL
	/* set local host if needed */
	if (loc_hostname != NULL) {
		if (P_dns4_name_to_IP(&loc_ip, loc_hostname) != 0)
			return P_net_err_dns;

		sin.sin_addr.s_addr = loc_ip;
	}

	/* set local port if needed */
	if (loc_port > 0)
		sin.sin_port = htons(loc_port);

	/* bind if anything local was set */
	if ((loc_hostname != NULL) || (loc_port > 0)) {
		if (bind(*sock, (struct sockaddr *) &sin, sizeof(sin)) != 0)
			return P_net_err_bind;
	}

	/* set up remote host and port */

	sin.sin_addr.s_addr = ip;
	sin.sin_port = htons(rem_port);

	if ((connect(*sock, (struct sockaddr*) &sin, sizeof(sin)) != 0) && (!P_net_inprogress)) {
		P_net_close(*sock);
		return P_net_err_connect;
	}

	/* Switch to non-blocking, if needed */
	if (non_blocking & P_net_nonblock_socket) {
		if (P_net_set_nonblocking(*sock) != 0 && (non_blocking & P_net_nonblock_quit_on_error)) {
			P_net_close(*sock);
			return P_net_err_nonblock;
		}
	}

	return 0;
}

int P_udp4_recvfrom(P_net_socket sock, char *buff, int maxlen, unsigned int *rem_ip, char **rem_host, int *rem_port)\
{
	P_ipv4_addr_t ip;
	struct sockaddr_in sin;
	int len, alen;

	alen = sizeof(sin);
	len = recvfrom(sock, buff, maxlen, 0, (struct sockaddr *)&sin, &alen);
	if (len < 0)
		return len;

	if (rem_ip != NULL)
		*rem_ip = ntohl(sin.sin_addr.s_addr);
	if (rem_port != NULL)
		*rem_port = ntohs(sin.sin_port);
	if (rem_host != NULL) {
		P_ipv4_addr_t ip;
		ip = ntohl(sin.sin_addr.s_addr);
		P_dns4_IP_to_name(rem_host, &ip);
	}
	return len;
}

