/*
    libporty - dynamic, splitting fifo
    Copyright (C) 2010..2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Contact (email and IRC): http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <libporty/config.h>
#include "buf.h"
#include "str.h"

#define GROW 256

static char *strclone(const char *str)
{
	int len;
	char *s;

	len = strlen(str);
	s   = malloc(len+1);
	memcpy(s, str, len+1);
	return s;
}

P_buf_t *P_buf_init(const char *sep)
{
	P_buf_t *b;
	b = malloc(sizeof(P_buf_t));
	b->alloced = 0;
	b->used    = 0;
	b->sep     = sep == NULL ? NULL : strclone(sep);
	b->next    = sep == NULL ? NULL : P_buf_init(NULL);
	b->last    = b->next;
	b->prev    = NULL;
	return b;
}

/* Append binary at the end of the buffer; a \0 at the new end will always
   be appended so current buffer is always a valid string, even if partial
   (does not have any use for real binary buffering)
   Also grows buffer if needed.
*/
static void append(P_buf_t **b, const char *in, int len)
{
	if ((*b)->used + len + 1 >= (*b)->alloced) {
		(*b)->alloced = (*b)->used + len + GROW;
		(*b) = realloc(*b, (*b)->alloced + sizeof(P_buf_t));
	}
	memcpy((*b)->buf+(*b)->used, in, len);
	(*b)->used += len;
	(*b)->buf[(*b)->used] = '\0';
}

void P_buf_fillbin(P_buf_t *buff, const char *data, int len_)
{
	const char *end, *start, *data_end;
	int len;
	P_buf_t *last, *nw, *prev;
	data_end = data+len_;

	for(end = start = data; end != NULL; ) {
		if (len_ > 0)
			end = P_binpbrk(start, data_end - start, buff->sep);
		else
			end = strpbrk(start, buff->sep);

		if (end == NULL)
			if (len_ > 0)
				len = data_end - start;
			else
				len = strlen(start);
		else
			len = end - start;

		prev = buff->last->prev;
		append(&(buff->last), start, len);
		if (prev != NULL)
			prev->next = buff->last;
		else
			buff->next = buff->last;

		if (end != NULL) {
			last = buff->last;
			nw = P_buf_init(NULL);

			last->next = nw;
			nw->prev = last;
			buff->last = nw;
		}
		start = end+1;
	}
}

void P_buf_fill(P_buf_t *buff, const char *strin)
{
	P_buf_fillbin(buff, strin, -1);
}

/* return next entry in the buffer (if return_partial == 0, return full
   entries only) or NULL. */
static P_inline P_buf_t *P_buf_get_(P_buf_t *buff, int return_partial)
{
	P_buf_t *b;
	b = buff->next;
	if ((b->next == NULL) && (!return_partial))
		return NULL;
	buff->next = b->next;
	if (b->next != NULL)
		b->next->prev = NULL;
	b->next = NULL;
	b->prev = NULL;

	return b;
}

char *P_buf_get(P_buf_t *buff)
{
	P_buf_t *b;
	b = P_buf_get_(buff, 0);
	if (b == NULL)
		return NULL;
	return b->buf;
}

char *P_buf_getbin(P_buf_t *buff, int *len)
{
	P_buf_t *b;
	b = P_buf_get_(buff, 0);
	if (b == NULL) {
		*len = -1;
		return NULL;
	}
	*len = b->used;
	return b->buf;
}

/* return the first N bytes wihtout any modification */
int P_buf_peekbin(P_buf_t *buff, char *res, int maxlen, int multirec)
{
	P_buf_t *b;
	int len, ofs;

	b = buff->next;
	for(len = 0, ofs = 0; len < maxlen; len++,ofs++) {
		if (ofs >= b->used) {
			if (!multirec)
				return len;

			b = b->next;
			if (b == NULL)
				return len;

			res[len] = buff->sep[0];
			len++;
			if (len >= maxlen)
				return len;
			ofs = 0;
		}
		res[len] = b->buf[ofs];
	}
	return len;
}


void P_buf_free(char *str)
{
	P_buf_t *b, f;
	int offs;

	offs = (char *)&(f.buf) - (char *)&f;
	b = (P_buf_t *)(str - offs);
	assert(b->sep == NULL);
	assert(b->next == NULL);
	assert(b->prev == NULL);

	free(b);
}

int P_buf_get_current_len(P_buf_t *buff)
{
	P_buf_t *b;
	b = buff->last;
	return b->used;
}

void P_buf_uninit(P_buf_t *buff)
{
	P_buf_t *b;
	while(buff->next != NULL) {
		b = P_buf_get_(buff, 1);
		P_buf_free(b->buf);
	}
	if (buff->sep != NULL)
		free(buff->sep);
	free(buff);
}

